package com.wechi.adweb.bridge.google.ads.service;

import com.google.ads.googleads.lib.GoogleAdsClient;
import com.google.ads.googleads.v18.services.GoogleAdsRow;
import com.google.ads.googleads.v18.services.GoogleAdsServiceClient;
import com.google.ads.googleads.v18.services.GoogleAdsServiceClient.SearchPagedResponse;
import com.google.ads.googleads.v18.services.SearchGoogleAdsRequest;
import com.google.api.core.ApiFuture;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.ads.common.CountryCode;
import com.wechi.adweb.bridge.google.ads.common.PlacementUtils;
import com.wechi.adweb.bridge.google.ads.dto.CampaignStatsDTO;
import com.wechi.adweb.bridge.google.ads.dto.CustomerStatsDTO;
import com.wechi.adweb.bridge.google.ads.dto.MetricsDTO;
import com.wechi.adweb.bridge.google.ads.dto.ReportRequestDTO;

import lombok.extern.slf4j.Slf4j;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

/**
 * @author wfansh
 */
@Slf4j
@Service
public class GoogleAdsService {
    @Autowired private GoogleAdsClient googleAdsClient;

    public CustomerStatsDTO getCustomerStats(String customerId) throws DataException {
        try (GoogleAdsServiceClient googleAdsServiceClient =
                googleAdsClient.getLatestVersion().createGoogleAdsServiceClient()) {

            // 1. Queries customer resource fields, conversions, total cost and optimization score.
            String customerQuery =
                    "SELECT customer.descriptive_name, customer.currency_code, customer.optimization_score, "
                            + "metrics.conversions, metrics.cost_micros "
                            + "FROM customer";
            ApiFuture<SearchPagedResponse> customerResponse =
                    googleAdsServiceClient
                            .searchPagedCallable()
                            .futureCall(
                                    SearchGoogleAdsRequest.newBuilder()
                                            .setCustomerId(customerId)
                                            .setQuery(customerQuery)
                                            .build());

            // 2. Queries customer balance via account budgets.
            String balanceQuery =
                    String.format(
                            "SELECT account_budget.adjusted_spending_limit_micros, account_budget.total_adjustments_micros, "
                                    + "account_budget.amount_served_micros "
                                    + "FROM account_budget "
                                    + "WHERE account_budget.status = 'APPROVED' AND account_budget.approved_end_date_time >= '%s'",
                            LocalDate.now());
            ApiFuture<SearchPagedResponse> balanceResponse =
                    googleAdsServiceClient
                            .searchPagedCallable()
                            .futureCall(
                                    SearchGoogleAdsRequest.newBuilder()
                                            .setCustomerId(customerId)
                                            .setQuery(balanceQuery)
                                            .build());

            // Waits for the completion of the above ApiFuture calls, using get() method.
            GoogleAdsRow customerResult = customerResponse.get().iterateAll().iterator().next();
            long balanceMicros =
                    StreamSupport.stream(balanceResponse.get().iterateAll().spliterator(), false)
                            .map(GoogleAdsRow::getAccountBudget)
                            .mapToLong(
                                    accountBudget ->
                                            accountBudget.getAdjustedSpendingLimitMicros()
                                                    - accountBudget.getAmountServedMicros())
                            .sum();

            return CustomerStatsDTO.builder()
                    .customerId(customerId)
                    .descriptiveName(customerResult.getCustomer().getDescriptiveName())
                    .currencyCode(customerResult.getCustomer().getCurrencyCode())
                    .conversions(customerResult.getMetrics().getConversions())
                    .costMicros(customerResult.getMetrics().getCostMicros())
                    // Balance.
                    .balanceMicros(balanceMicros)
                    // Optimization score.
                    .optiScore(customerResult.getCustomer().getOptimizationScore())
                    .build();
        } catch (InterruptedException | ExecutionException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public Map<String, MetricsDTO> getDailyStats(ReportRequestDTO reportRequest)
            throws DataException {
        try (GoogleAdsServiceClient googleAdsServiceClient =
                googleAdsClient.getLatestVersion().createGoogleAdsServiceClient()) {
            String query =
                    String.format(
                            "SELECT customer.descriptive_name, customer.currency_code, "
                                    + "segments.date, "
                                    + "metrics.impressions, metrics.clicks, metrics.ctr, metrics.average_cpc, "
                                    + "metrics.average_cpm, metrics.conversions, metrics.cost_micros "
                                    + "FROM customer WHERE %s "
                                    + "ORDER BY segments.date LIMIT %d",
                            reportRequest.toDateClause(), reportRequest.getLimit());

            SearchGoogleAdsRequest request =
                    SearchGoogleAdsRequest.newBuilder()
                            .setCustomerId(reportRequest.getCustomerId())
                            .setQuery(query)
                            .build();

            SearchPagedResponse response = googleAdsServiceClient.search(request);
            return StreamSupport.stream(response.iterateAll().spliterator(), false)
                    .collect(
                            Collectors.toMap(
                                    googleAdsRow -> googleAdsRow.getSegments().getDate(),
                                    googleAdsRow ->
                                            MetricsDTO.fromMetrics(googleAdsRow.getMetrics()),
                                    // For duplicated dates.
                                    MetricsDTO::merge,
                                    // Preserves the original order in the response.
                                    LinkedHashMap::new));
        }
    }

    public List<CampaignStatsDTO> getCampaignStats(ReportRequestDTO reportRequest)
            throws DataException {
        try (GoogleAdsServiceClient googleAdsServiceClient =
                googleAdsClient.getLatestVersion().createGoogleAdsServiceClient()) {
            String query =
                    String.format(
                            "SELECT campaign.name, campaign.status, campaign.primary_status, campaign.bidding_strategy_type, "
                                    + "campaign.advertising_channel_type, campaign.advertising_channel_sub_type, campaign_budget.amount_micros, "
                                    + "metrics.impressions, metrics.clicks, metrics.ctr, metrics.average_cpc, "
                                    + "metrics.average_cpm, metrics.conversions, metrics.cost_micros "
                                    + "FROM campaign "
                                    + "WHERE campaign.status != 'REMOVED' AND %s "
                                    + "ORDER BY metrics.impressions DESC, metrics.clicks DESC LIMIT %d",
                            reportRequest.toDateClause(), reportRequest.getLimit());

            SearchGoogleAdsRequest request =
                    SearchGoogleAdsRequest.newBuilder()
                            .setCustomerId(reportRequest.getCustomerId())
                            .setQuery(query)
                            .build();

            SearchPagedResponse response = googleAdsServiceClient.search(request);
            return StreamSupport.stream(response.iterateAll().spliterator(), false)
                    .map(CampaignStatsDTO::fromGoogleAdsRow)
                    .toList();
        }
    }

    public Map<String, MetricsDTO> getKeywordStats(ReportRequestDTO reportRequest)
            throws DataException {
        try (GoogleAdsServiceClient googleAdsServiceClient =
                googleAdsClient.getLatestVersion().createGoogleAdsServiceClient()) {
            String query =
                    String.format(
                            "SELECT ad_group_criterion.keyword.text, ad_group.name, campaign.name, "
                                    + "metrics.impressions, metrics.clicks, metrics.ctr, metrics.average_cpc, "
                                    + "metrics.average_cpm, metrics.conversions, metrics.cost_micros "
                                    + "FROM keyword_view "
                                    + "WHERE %s "
                                    + "ORDER BY metrics.impressions DESC, metrics.clicks DESC LIMIT %d",
                            reportRequest.toDateClause(), reportRequest.getLimit());

            SearchGoogleAdsRequest request =
                    SearchGoogleAdsRequest.newBuilder()
                            .setCustomerId(reportRequest.getCustomerId())
                            .setQuery(query)
                            .build();

            SearchPagedResponse response = googleAdsServiceClient.search(request);
            return StreamSupport.stream(response.iterateAll().spliterator(), false)
                    .collect(
                            Collectors.toMap(
                                    googleAdsRow ->
                                            googleAdsRow
                                                    .getAdGroupCriterion()
                                                    .getKeyword()
                                                    .getText(),
                                    googleAdsRow ->
                                            MetricsDTO.fromMetrics(googleAdsRow.getMetrics()),
                                    // For duplicated keywords.
                                    MetricsDTO::merge,
                                    // Preserves the original order in the response.
                                    LinkedHashMap::new));
        }
    }

    /**
     * Uses 'group_placement_view' instead of 'detail_placement_view', corresponding to the "WHERE
     * ADS SHOWED" menu on the UI.
     *
     * <p>See https://groups.google.com/g/adwords-api/c/2sdF9iuBns0/m/6gzfhd2HAwAJ for details.
     */
    public Map<String, MetricsDTO> getPlacementStats(ReportRequestDTO reportRequest)
            throws DataException {
        try (GoogleAdsServiceClient googleAdsServiceClient =
                googleAdsClient.getLatestVersion().createGoogleAdsServiceClient()) {
            String query =
                    String.format(
                            "SELECT group_placement_view.display_name, group_placement_view.placement, "
                                    + "group_placement_view.placement_type, group_placement_view.target_url, "
                                    + "metrics.impressions, metrics.clicks, metrics.ctr, metrics.average_cpc, "
                                    + "metrics.average_cpm, metrics.conversions, metrics.cost_micros "
                                    + "FROM group_placement_view "
                                    + "WHERE %s "
                                    + "ORDER BY metrics.impressions DESC, metrics.clicks DESC LIMIT %d",
                            reportRequest.toDateClause(), reportRequest.getLimit());

            SearchGoogleAdsRequest request =
                    SearchGoogleAdsRequest.newBuilder()
                            .setCustomerId(reportRequest.getCustomerId())
                            .setQuery(query)
                            .build();

            SearchPagedResponse response = googleAdsServiceClient.search(request);
            return StreamSupport.stream(response.iterateAll().spliterator(), false)
                    .collect(
                            Collectors.toMap(
                                    googleAdsRow ->
                                            PlacementUtils.format(
                                                    googleAdsRow.getGroupPlacementView()),
                                    googleAdsRow ->
                                            MetricsDTO.fromMetrics(googleAdsRow.getMetrics()),
                                    // For duplicated placements.
                                    MetricsDTO::merge,
                                    // Preserves the original order in the response.
                                    LinkedHashMap::new));
        }
    }

    public Map<String, MetricsDTO> getCountryStats(ReportRequestDTO reportRequest)
            throws DataException {
        try (GoogleAdsServiceClient googleAdsServiceClient =
                googleAdsClient.getLatestVersion().createGoogleAdsServiceClient()) {
            String query =
                    String.format(
                            "SELECT geographic_view.country_criterion_id, geographic_view.location_type, "
                                    + "metrics.impressions, metrics.clicks, metrics.ctr, metrics.average_cpc, "
                                    + "metrics.average_cpm, metrics.conversions, metrics.cost_micros "
                                    + "FROM geographic_view "
                                    + "WHERE %s "
                                    + "ORDER BY metrics.impressions DESC, metrics.clicks DESC LIMIT %d",
                            reportRequest.toDateClause(), reportRequest.getLimit());

            SearchGoogleAdsRequest request =
                    SearchGoogleAdsRequest.newBuilder()
                            .setCustomerId(reportRequest.getCustomerId())
                            .setQuery(query)
                            .build();

            SearchPagedResponse response = googleAdsServiceClient.search(request);
            return StreamSupport.stream(response.iterateAll().spliterator(), false)
                    .collect(
                            Collectors.toMap(
                                    googleAdsRow ->
                                            CountryCode.valueOfId(
                                                            googleAdsRow
                                                                    .getGeographicView()
                                                                    .getCountryCriterionId())
                                                    .getCode(),
                                    googleAdsRow ->
                                            MetricsDTO.fromMetrics(googleAdsRow.getMetrics()),
                                    // For duplicated countries.
                                    MetricsDTO::merge,
                                    // Preserves the original order in the response.
                                    LinkedHashMap::new));
        }
    }
}
